import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import HexModelGrid, RasterModelGrid
from landlab.components import NormalFault


def test_dx_equals_zero():
    """Test a vertical fault trace."""
    grid = RasterModelGrid((6, 6), xy_spacing=10)

    grid.add_zeros("topographic__elevation", at="node")

    param_dict = {
        "faulted_surface": "topographic__elevation",
        "fault_dip_angle": 90.0,
        "fault_throw_rate_through_time": {"time": [0, 9, 10], "rate": [0, 0, 0.05]},
        "fault_trace": {"y1": 0, "x1": 30, "y2": 30, "x2": 30},
        "include_boundaries": True,
    }

    nf = NormalFault(grid, **param_dict)

    out = np.array(
        [
            [True, True, True, False, False, False],
            [True, True, True, False, False, False],
            [True, True, True, False, False, False],
            [True, True, True, False, False, False],
            [True, True, True, False, False, False],
            [True, True, True, False, False, False],
        ],
        dtype=bool,
    )

    assert_array_equal(nf.faulted_nodes.reshape(grid.shape), out)


def test_anti_aximuth_greq_2pi():
    """Test anti azimuth over 2*pi."""
    grid = RasterModelGrid((6, 6), xy_spacing=10)

    grid.add_zeros("topographic__elevation", at="node")

    param_dict = {
        "faulted_surface": "topographic__elevation",
        "fault_dip_angle": 90.0,
        "fault_throw_rate_through_time": {"time": [0, 9, 10], "rate": [0, 0, 0.05]},
        "fault_trace": {"y1": 30.0, "x1": 30.0, "y2": 20.0, "x2": 0.0},
        "include_boundaries": True,
    }

    nf = NormalFault(grid, **param_dict)

    assert nf._fault_anti_azimuth > 2.0 * np.pi

    out = np.array(
        [
            [True, True, True, True, True, True],
            [True, True, True, True, True, True],
            [True, True, True, True, True, True],
            [False, False, False, False, True, True],
            [False, False, False, False, False, False],
            [False, False, False, False, False, False],
        ],
        dtype=bool,
    )

    assert_array_equal(nf.faulted_nodes.reshape(grid.shape), out)


def test_non_raster():
    """Test a hex model grid."""
    grid = HexModelGrid((7, 3), spacing=10, xy_of_lower_left=(-15.0, 0.0))

    grid.add_zeros("topographic__elevation", at="node")

    param_dict = {
        "faulted_surface": "topographic__elevation",
        "fault_dip_angle": 90.0,
        "fault_throw_rate_through_time": {"time": [0, 9, 10], "rate": [0, 0, 0.05]},
        "fault_trace": {"y1": 30.0, "x1": 30.0, "y2": 20.0, "x2": 0.0},
        "include_boundaries": True,
    }

    nf = NormalFault(grid, **param_dict)

    # plotting, to test this. it works!
    # import matplotlib.pyplot as plt
    # plt.figure()
    # imshow_grid(grid, nf.faulted_nodes, color_for_background='y')
    # plt.plot(grid.x_of_node, grid.y_of_node, 'c.')
    # plt.plot([param_dict['fault_trace']['x1'], param_dict['fault_trace']['x2']],
    #         [param_dict['fault_trace']['y1'], param_dict['fault_trace']['y2']], 'r')
    # plt.show()

    out = np.array(
        [
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ],
        dtype=bool,
    )

    assert_array_equal(nf.faulted_nodes, out)


def test_dip_geq_90():
    """Test dip angles of >90 degrees."""
    grid = RasterModelGrid((6, 6), xy_spacing=10)

    grid.add_zeros("topographic__elevation", at="node")

    with pytest.raises(ValueError):
        NormalFault(grid, fault_dip_angle=90.001)


def test_uplifting_multiple_fields():
    """Test uplifting multiple fields with NormalFault."""
    grid = RasterModelGrid((6, 6), xy_spacing=10)

    grid.add_zeros("topographic__elevation", at="node")

    zbr = grid.add_zeros("bedrock__elevation", at="node")

    zbr -= 1.0

    param_dict = {
        "faulted_surface": ["topographic__elevation", "bedrock__elevation"],
        "fault_dip_angle": 90.0,
        "fault_throw_rate_through_time": {"time": [0, 9, 10], "rate": [1, 1, 1]},
        "fault_trace": {"y1": 30.0, "x1": 30.0, "y2": 20.0, "x2": 0.0},
        "include_boundaries": True,
    }

    nf = NormalFault(grid, **param_dict)

    nf.run_one_step(dt=10)

    elev = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            10.0,
            10.0,
            10.0,
            10.0,
            0.0,
            0.0,
            10.0,
            10.0,
            10.0,
            10.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            10.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    bedrock = np.array(
        [
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            9.0,
            9.0,
            9.0,
            9.0,
            -1.0,
            -1.0,
            9.0,
            9.0,
            9.0,
            9.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            9.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
        ]
    )

    assert_array_equal(grid.at_node["topographic__elevation"], elev)
    assert_array_equal(grid.at_node["bedrock__elevation"], bedrock)


def test_uplifting_a_not_yet_created_field():
    """Test uplifting a field that does not exist with  NormalFault."""
    grid = RasterModelGrid((6, 6), xy_spacing=10)

    grid.add_zeros("topographic__elevation", at="node")

    zbr = grid.add_zeros("bedrock__elevation", at="node")

    zbr -= 1.0

    param_dict = {
        "faulted_surface": [
            "topographic__elevation",
            "bedrock__elevation",
            "spam",
            "eggs",
        ],
        "fault_dip_angle": 90.0,
        "fault_throw_rate_through_time": {"time": [0, 9, 10], "rate": [1, 1, 1]},
        "fault_trace": {"y1": 30.0, "x1": 30.0, "y2": 20.0, "x2": 0.0},
        "include_boundaries": True,
    }

    assert "spam" not in grid.at_node
    assert "eggs" not in grid.at_node

    # instantiating NormalFault will not create spam or eggs
    nf = NormalFault(grid, **param_dict)

    assert "spam" not in grid.at_node
    assert "eggs" not in grid.at_node

    assert "spam" in nf._not_yet_instantiated
    assert "eggs" in nf._not_yet_instantiated

    # running NormalFault will not create spam or eggs
    nf.run_one_step(dt=10)

    assert "spam" not in grid.at_node
    assert "eggs" not in grid.at_node

    assert "spam" in nf._not_yet_instantiated
    assert "eggs" in nf._not_yet_instantiated

    # running NormalFault after adding spam and eggs will result in NormalFault
    # modifying these fields.

    grid.add_zeros("eggs", at="node")

    grid.add_zeros("spam", at="node")

    nf.run_one_step(dt=10)

    assert "spam" in grid.at_node
    assert "eggs" in grid.at_node

    vals = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            10.0,
            10.0,
            10.0,
            10.0,
            0.0,
            0.0,
            10.0,
            10.0,
            10.0,
            10.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            10.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert_array_equal(grid.at_node["eggs"], vals)
    assert_array_equal(grid.at_node["spam"], vals)



""" Implementation of the FramedVoronoiGraph and its static layout:
HorizontalRectVoronoiGraph. This pattern is inspired from the developments of the HexModelGrid

.. codeauthor:: sebastien lenard
"""

from functools import cached_property

import numpy as np

from ...utils.decorators import make_return_array_immutable
from ..graph import Graph
from ..voronoi.voronoi import DelaunayGraph


class HorizontalRectVoronoiGraph:
    """The horizontal rectangular frame for the FramedVoronoiGraph."""

    @staticmethod
    def number_of_nodes(shape):
        """
        Parameters
        ----------
        shape: tuple of int
            Number of rows and number of columns

        Returns
        -------
        int
            Number of nodes

        Examples
        --------
        >>> from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph
        >>> HorizontalRectVoronoiGraph.number_of_nodes((3, 2))
        6
        """
        n_rows, n_cols = shape
        return n_rows * n_cols

    @staticmethod
    def xy_of_node(
        shape,
        xy_spacing=(1.0, 1.0),
        xy_of_lower_left=(0.0, 0.0),
        xy_min_spacing=(0.5, 0.5),
        seed=200,
    ):
        """The x and y coordinates of the graph's nodes.

        Calculation of the x-y coordinates is done following these steps:

        1. Generate a rectangular, regular meshgrid.
        2. Move the coordinates of the core nodes over a random distance around their
           initial position, within a threshold calculated from *xy_spacing* and
           *xy_min_spacing*.
        3. Rectify the y-coordinates of the nodes of the left and right to ensure
           that the leftmost node of a row has a lower y than the rightmost node.
           This ensures that the ids of these nodes are not modified by subsequent
           sorting operations on the graph and make it possible to get the
           perimeter nodes in simple way.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of nodes.
        xy_spacing : float or tuple of float, optional
            Node spacing along x and y coordinates. If ``float``, same spacing at *x* and *y*.
        xy_of_lower_left : tuple, optional
            Minimum *x*-of-node and *y*-of-node values. Depending on the grid,
            a node may not be present at this location.
        xy_min_spacing: float or tuple of float, optional
            Final minimal spacing between nodes. Random moves of the core nodes
            around their initial positions cannot be above this threshold:
            ``(xy_spacing - xy_min_spacing) / 2``.  If ``float``, same minimal
            spacing for *x* and *y*.
        seed: int, optional
            Seed used to generate the random *x* and *y* moves. When set, controls a
            pseudo-randomness of moves to ensure reproducibility.
            When ``None``, seed is random and the moves of coordinates are
            completely random.

        Returns
        -------
        x_of_node, y_of_node : ndarray of float
            The arrays of *x* and *y* coordinates.

        Examples
        --------
        >>> from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph

        >>> x_of_node, y_of_node = HorizontalRectVoronoiGraph.xy_of_node((3, 3), seed=200)

        Coordinates of the lower left node,

        >>> x_of_node[0], y_of_node[0]
        (0.0, 0.0)

        *x* coordinates of the left and right edges,

        >>> x_of_node[3], x_of_node[5]
        (0.0, 2.0)

        *y* coordinate of the middle row of the left edge,

        >>> y_of_node[3]
        0.749
        """

        n_rows, n_cols = shape
        max_move = (
            (xy_spacing[0] - xy_min_spacing[0]) / 2,
            (xy_spacing[1] - xy_min_spacing[1]) / 2,
        )

        if max_move[0] < 0.0 or max_move[1] < 0.0:
            raise ValueError("minimum spacing must be greater than node spacing")
        if np.allclose(max_move, 0.0):
            raise ValueError("at least one of x and y moves must be greater than zero")

        # Generation of a rectangular grid, coordinates must be float
        x_of_node, y_of_node = np.meshgrid(
            np.arange(n_cols, dtype=float) * xy_spacing[0] + xy_of_lower_left[0],
            np.arange(n_rows, dtype=float) * xy_spacing[1] + xy_of_lower_left[1],
        )
        # Randomly move the coordinates of the core nodes of the grid. Move
        # below +/- (spacing - min_spacing) / 2
        xy_random_generator = np.random.default_rng(seed=seed)

        x_moves = xy_random_generator.uniform(-max_move[0], max_move[0], shape)
        y_moves = xy_random_generator.uniform(-max_move[1], max_move[1], shape)

        x_of_node[1:-1, 1:-1] += x_moves[1:-1, 1:-1]
        y_of_node[1:-1, 1:-1] += y_moves[1:-1, 1:-1]
        # Control the node id attribution for left and right edge. For instance, for a 3x3 grid,
        # make sure that node 3 is at the left of the 2nd row and node 5 at the right.
        # For this, for each core row, set y of the leftmost node as the minimal y of the row
        # and set y of the rightmost node as the maximal y of the row
        for i in range(1, n_rows - 1):
            y_of_node[i, 0] -= max_move[1] + 1.0e-3
            y_of_node[i, n_cols - 1] += max_move[1] + 1.0e-3

        return x_of_node.reshape(-1), y_of_node.reshape(-1)

    @staticmethod
    def corner_nodes(shape):
        """
        Parameters
        ----------
        shape: tuple of int
            Number of rows and number of columns

        Returns
        -------
        ndarray of int
            Ids of the corner nodes

        Examples
        --------
        >>> from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph
        >>> HorizontalRectVoronoiGraph.corner_nodes((3, 4))
        (11, 8, 0, 3)
        """
        n_rows, n_cols = shape
        return (n_rows * n_cols - 1, n_cols * (n_rows - 1), 0, n_cols - 1)

    @staticmethod
    def number_of_perimeter_nodes(shape):
        """
        Parameters
        ----------
        shape: tuple of int
            Number of rows and number of columns

        Returns
        -------
        int
            Number of perimeter nodes

        Examples
        --------
        >>> from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph
        >>> HorizontalRectVoronoiGraph.number_of_perimeter_nodes((3, 4))
        10
        """
        if 1 in shape:
            return np.prod(shape)
        return 2 * shape[0] + 2 * (shape[1] - 2)

    @staticmethod
    def perimeter_nodes(shape):
        """
        Parameters
        ----------
        shape: tuple of int
            Number of rows and number of columns

        Returns
        -------
        ndarray of int
            Ids of the perimeter nodes

        Examples
        --------
        >>> from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph
        >>> HorizontalRectVoronoiGraph.perimeter_nodes((3, 3))
        array([2, 5, 8, 7, 6, 3, 0, 1])
        """
        return np.concatenate(HorizontalRectVoronoiGraph.nodes_at_edge(shape))

    @staticmethod
    def nodes_at_edge(shape):
        """
        Parameters
        ----------
        shape: tuple of int
            Number of rows and number of columns

        Returns
        -------
        right, top, left, bottom : ndarray of int
            For each edge give the ids of the nodes present at the edge

        Examples
        --------
        >>> from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph
        >>> HorizontalRectVoronoiGraph.nodes_at_edge((3, 3))
        (array([2, 5]), array([8, 7]), array([6, 3]), array([0, 1]))
        """
        n_rows, n_cols = shape
        if n_rows == n_cols == 1:
            return (np.array([0]),) + (np.array([], dtype=int),) * 3
        (
            northeast,
            northwest,
            southwest,
            southeast,
        ) = HorizontalRectVoronoiGraph.corner_nodes(shape)

        if n_rows > 1:
            south = np.arange(southwest, southeast)
        else:
            south = np.array([southwest], dtype=int)

        if n_cols > 1:
            west = np.arange(northwest, southwest, -n_cols)
        else:
            west = np.array([northwest], dtype=int)

        return (
            np.arange(southeast, northeast, n_cols),
            np.arange(northeast, northwest, -1),
            west,
            south,
        )


class FramedVoronoiGraph(DelaunayGraph):
    """VoronoiDelaunay graph based on a fixed lattice.

    Graph of an unstructured grid of Voronoi Delaunay cells and
    irregular patches. It is a special type of :class`~.VoronoiDelaunayGraph` in which
    the initial set of points is arranged in a fixed lattice (e.g. like a
    :class:`~.RasterModelGrid`) named here "layout" and the core points are
    then moved from their initial position by a random distance, lower than a
    certain threshold.

    Examples
    --------
    >>> from landlab.graph import FramedVoronoiGraph

    >>> graph = FramedVoronoiGraph((3, 3), seed=200)
    >>> graph.number_of_nodes
    9

    >>> graph.x_of_node[2:4]    # doctest: +NORMALIZE_WHITESPACE
    array([ 2.,  0.])
    >>> graph.y_of_node[2:4]    # doctest: +NORMALIZE_WHITESPACE
    array([ 0.   ,  0.749])
    >>> graph.y_of_node[5]    # doctest: +NORMALIZE_WHITESPACE
    1.2509999999999999

    >>> graph.number_of_links
    16
    >>> graph.number_of_patches
    8
    """

    def __init__(
        self,
        shape,
        xy_spacing=(1.0, 1.0),
        xy_of_lower_left=(0.0, 0.0),
        sort=False,
        xy_min_spacing=(0.5, 0.5),
        seed=200,
    ):
        """Create the graph.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of nodes.
        xy_spacing : float or tuple of float, optional
            Node spacing along *x* and *y* coordinates. If ``float``, same spacing *x* and *y*
            spacing.
        xy_of_lower_left : tuple, optional
            Minimum *x*-of-node and *y*-of-node values. Depending on the grid,
            a node may not be present at this location.
        sort: bool
            If ``True``, nodes, links and patches are re-numbered according
            certain their positions.  Currently not used.
        xy_min_spacing: float or tuple of float, optional
            Final minimal spacing between nodes. Random moves of the core nodes
            around their position cannot be above this threshold:
            ``(xy_spacing - xy_min_spacing) / 2``
            If ``float``, same minimal spacing for *x* and *y*.
        seed: int, optional
            Seed used to generate the random *x* and *y* moves.
            When set, controls a pseudo-randomness of moves to ensure
            reproducibility. When ``None``, seed is random and the moves of coordinates are
            completely random.

        Returns
        -------
        FramedVoronoiGraph
            A newly-created graph.

        Examples
        --------
        Create a grid with 3 rows and 2 columns of nodes.

        >>> from landlab.graph import FramedVoronoiGraph
        >>> graph = FramedVoronoiGraph((3, 2))
        >>> graph.number_of_nodes
        6
        """
        # 1. Check and format input arguments
        #####################################
        self._shape = shape
        self._seed = seed

        try:
            xy_spacing = np.asfarray(np.broadcast_to(xy_spacing, 2))
        except TypeError as exc:
            raise TypeError("spacing must be a float or a tuple of floats") from exc
        else:
            self._xy_spacing = xy_spacing[0], xy_spacing[1]

        try:
            xy_of_lower_left = np.asfarray(np.broadcast_to(xy_of_lower_left, 2))
        except TypeError as exc:
            raise TypeError(
                "xy of lower left must be a float or a tuple of floats"
            ) from exc
        else:
            self._xy_of_lower_left = xy_of_lower_left[0], xy_of_lower_left[1]

        node_layout = self._node_layout = "rect"
        orientation = self._orientation = "horizontal"

        layouts = {
            "horizontal_rect": HorizontalRectVoronoiGraph,
        }
        layout = layouts["_".join([orientation, node_layout])]

        try:
            xy_min_spacing = np.asfarray(np.broadcast_to(xy_min_spacing, 2))
        except TypeError as exc:
            raise TypeError(
                "minimal spacing must be a float or a tuple of floats"
            ) from exc
        else:
            self._xy_min_spacing = xy_min_spacing[0], xy_min_spacing[1]

        # 2. Construction of the layout and the x-y coordinates of nodes
        ################################################################
        x_of_node, y_of_node = layout.xy_of_node(
            self._shape,
            xy_spacing=self._xy_spacing,
            xy_of_lower_left=self._xy_of_lower_left,
            xy_min_spacing=self._xy_min_spacing,
            seed=self._seed,
        )

        # 3. Determination of the perimeter and edge nodes
        #########################################
        self._perimeter_nodes = layout.perimeter_nodes(self._shape)

        (right, top, left, bottom) = layout.nodes_at_edge(self._shape)
        self._nodes_at_right_edge = np.sort(np.append(right, top[0]))
        self._nodes_at_top_edge = np.sort(np.append(top, left[0]))
        self._nodes_at_left_edge = np.sort(np.append(left, bottom[0]))
        self._nodes_at_bottom_edge = np.sort(np.append(bottom, right[0]))

        perimeter_links = np.empty((len(self._perimeter_nodes), 2), dtype=int)
        perimeter_links[:, 0] = self._perimeter_nodes
        perimeter_links[:-1, 1] = self._perimeter_nodes[1:]
        perimeter_links[-1, 1] = self._perimeter_nodes[0]

        self._x_of_node = x_of_node
        self._y_of_node = y_of_node
        self._perimeter_links = perimeter_links

        # 3. Instantiation of the parent class
        ######################################
        if 1 in shape:
            Graph.__init__(
                self,
                (y_of_node, x_of_node),
                links=list(
                    zip(np.arange(len(y_of_node) - 1), np.arange(1, len(y_of_node)))
                ),
                sort=False,
            )
        else:
            DelaunayGraph.__init__(
                self,
                (y_of_node, x_of_node),
                perimeter_links=perimeter_links,
                sort=False,
            )

    @property
    def shape(self):
        return self._shape

    @property
    def xy_spacing(self):
        return self._xy_spacing

    @property
    def orientation(self):
        return self._orientation

    @property
    def node_layout(self):
        return self._node_layout

    @cached_property
    @make_return_array_immutable
    def perimeter_nodes(self):
        return self._perimeter_nodes

    @property
    @make_return_array_immutable
    def nodes_at_right_edge(self):
        return self._nodes_at_right_edge

    @property
    @make_return_array_immutable
    def nodes_at_top_edge(self):
        return self._nodes_at_top_edge

    @property
    @make_return_array_immutable
    def nodes_at_left_edge(self):
        return self._nodes_at_left_edge

    @property
    @make_return_array_immutable
    def nodes_at_bottom_edge(self):
        return self._nodes_at_bottom_edge



import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.backends.backend_pdf import PdfPages

import landlab
from landlab.plot.imshow import _guess_location_from_name, _guess_location_from_size


@pytest.mark.slow
def test_imshow_grid():
    rmg = landlab.RasterModelGrid((4, 5))

    pp = PdfPages("test.pdf")

    values = np.arange(rmg.number_of_nodes)
    landlab.plot.imshow_grid(rmg, values, at="node", limits=(0, 20))
    pp.savefig()

    plt.clf()
    rmg.status_at_node[7] = rmg.BC_NODE_IS_CLOSED
    values = np.arange(rmg.number_of_cells)
    landlab.plot.imshow_grid(rmg, values, at="cell", symmetric_cbar=True)
    pp.savefig()
    pp.close()


def test_imshow_grid_input():
    rmg = landlab.RasterModelGrid((4, 5))
    values = np.arange(rmg.number_of_nodes - 1)
    with pytest.raises(ValueError):
        landlab.plot.imshow_grid(rmg, values, at="node", limits=(0, 20))


def test_imshowhs_grid_array_wrong_size():
    rmg = landlab.RasterModelGrid((4, 5))
    values = np.arange(rmg.number_of_nodes - 1)
    with pytest.raises(ValueError):
        landlab.plot.imshowhs_grid(rmg, values, at="node", limits=(0, 20))


def test_imshowhs_grid_array_wrong_size_with_field_name():
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    values1 = np.arange(mg.number_of_nodes - 1)

    with pytest.raises(ValueError):
        landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            drape1=values1,
            plot_type="Drape1",
            var_name="Soil",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            ticks_km=False,
            limits=(0, 2),
        )


def test_imshowhs_grid_array_size_mismatch():
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    values1 = np.arange(mg.number_of_nodes)
    values2 = np.arange(mg.number_of_nodes - 1)

    with pytest.raises(ValueError):
        landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            drape1=values1,
            drape2=values2,
            plot_type="Drape2",
            var_name="Soil",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            ticks_km=False,
            limits=(0, 2),
        )


def test_imshowhs_grid_1():
    """Show DEM draped over the shaded topographic relief"""
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        var_name="Topo",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        symmetric_cbar=True,
        limits=(0, 10),
    )


def test_imshowhs_grid_2():
    """Show DEM draped over the shaded topographic relief with exaggeration"""
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        var_name="Topo",
        var_units=r"m",
        grid_units=("m", "m"),
        vertical_exa=2,
        ticks_km=True,
        symmetric_cbar=True,
        vmin=0,
        vmax=10,
    )


def test_imshowhs_grid_3():
    """Show Hillshade"""
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        plot_type="Hillshade",
        var_name="Topo",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        plt_contour=True,
        vmax=10,
        vmin=0,
    )


def test_imshowhs_grid_4a():
    """
    Show Drape1 draped over the shaded topographic relief with exaggeration
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    # Show Soil thickness draped over the shaded topographic relief
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["soil__depth"],
        plot_type="Drape1",
        var_name="Soil",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        limits=(0, 2),
    )


def test_imshowhs_grid_4b():
    """
    Show Drape1 draped over the shaded topographic relief with exaggeration
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    # Show Soil thickness draped over the shaded topographic relief
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["soil__depth"],
        plot_type="Drape1",
        var_name="Soil",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        vmin=0,
        vmax=2,
        plt_contour=True,
    )


def test_imshowhs_grid_4c():
    """
    Show Drape1 draped over the shaded topographic relief with exaggeration
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    # Show Soil thickness draped over the shaded topographic relief
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["soil__depth"],
        plot_type="Drape1",
        var_name="Soil",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        symmetric_cbar=True,
    )


def test_imshowhs_grid_5():
    """
    Show Drape1 draped over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_zeros("Layer_1", at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        plot_type="Drape1",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
    )


def test_imshowhs_grid_6a():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_full("Layer_1", 10.0, at="node")
    mg.add_full("Layer_2", 100.0, at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        limits=(0, 200),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="white",
    )


def test_imshowhs_grid_6b():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <10
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_full("Layer_1", 10.0, at="node")
    mg.add_full("Layer_2", 100.0, at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="white",
        vmin=0,
        vmax=9,
    )


def test_imshowhs_grid_6c():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <100
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_full("Layer_1", 10.0, at="node")
    mg.add_full("Layer_2", 100.0, at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="white",
        vmin=0,
        vmax=99,
    )


def test_imshowhs_grid_6d():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <100
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_full("Layer_1", 10.0, at="node")
    mg.add_full("Layer_2", 100.0, at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="white",
        vmin=0,
        vmax=999,
    )


def test_imshowhs_grid_6e():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <100
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_full("Layer_1", 10.0, at="node")
    mg.add_full("Layer_2", 100.0, at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="white",
        add_double_colorbar=True,
        vmin=0,
        vmax=99999,
    )


def test_imshowhs_grid_7():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_zeros("Layer_1", at="node")
    mg.add_zeros("Layer_2", at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1="topographic__elevation",
        drape2="soil__depth",
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="white",
        thres_drape2=1,
        cmap2=None,
        add_double_colorbar=True,
    )


def test_imshowhs_grid_8():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax >10<100
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_full("Layer_1", 10.0, at="node")
    mg.add_full("Layer_2", 100.0, at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1="topographic__elevation",
        drape2="soil__depth",
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="white",
        thres_drape2=1,
        cmap2=None,
        add_double_colorbar=True,
        vmin=0,
        vmax=99,
    )


def test_imshowhs_grid_9():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax>100
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_zeros("Layer_1", at="node")
    mg.add_zeros("Layer_2", at="node")
    landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1="topographic__elevation",
        drape2="soil__depth",
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        ticks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed=None,
        thres_drape2=1,
        cmap2=None,
        add_double_colorbar=True,
        vmin=0,
        vmax=99999,
    )
    with pytest.raises(ValueError):
        landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            plot_type="Oops",
        )


def test_imshowhs_grid_10():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_zeros("Layer_1", at="node")
    mg.add_zeros("Layer_2", at="node")
    with pytest.raises(ValueError):
        landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            drape1=mg.at_node["Layer_1"],
            plot_type="Drape2",
            var_name="Layer 1",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            ticks_km=False,
            limits=(0, 2),
            colorbar_label_y=-55,
            add_label_bbox=True,
            thres_drape1=0.001,
        )


def test_imshowhs_grid_11():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_zeros("Layer_1", at="node")
    mg.add_zeros("Layer_2", at="node")
    with pytest.raises(ValueError):
        landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            plot_type="Drape1",
            var_name="Layer 1",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            ticks_km=False,
            limits=(0, 2),
            colorbar_label_y=-55,
            add_label_bbox=True,
            thres_drape1=0.001,
        )


@pytest.mark.parametrize("valid_units", [None, "foo"])
def test_imshowhs_var_units(valid_units):
    """units should be a string or None"""
    mg = landlab.RasterModelGrid((4, 5))
    mg.add_zeros("topographic__elevation", at="node")
    landlab.plot.imshowhs_grid(mg, "topographic__elevation", var_units=valid_units)


def test_hex_grid_not_allowed():
    """Currently no support for hex"""
    mg = landlab.HexModelGrid((5, 3))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(NotImplementedError):
        landlab.plot.imshowhs_grid(mg, "topographic__elevation")


@pytest.mark.parametrize("at", ["link", "patch", "corner", "face", "cell"])
def test_at_anything_but_node(at):
    """Currently no support for anything but node."""
    mg = landlab.RasterModelGrid((5, 3))
    mg.add_empty("topographic__elevation", at=at)
    with pytest.raises(NotImplementedError):
        landlab.plot.imshowhs_grid(mg, "topographic__elevation", at=at)


@pytest.mark.parametrize("at", ["node", "cell"])
def test_imshow_grid_guess_from_name(at):
    grid = landlab.RasterModelGrid((3, 4))
    grid.add_zeros("z", at=at)
    landlab.plot.imshow_grid(grid, "z")


@pytest.mark.parametrize("at", ["node", "cell"])
def test_imshow_grid_guess_from_size(at):
    grid = landlab.RasterModelGrid((3, 4))
    values = grid.zeros(at=at)
    landlab.plot.imshow_grid(grid, values)


def test_imshow_grid_unknown_location():
    expected = "unable to determine location of values, use 'at' keyword"
    grid = landlab.RasterModelGrid((5, 3))
    values = np.empty(grid.number_of_links + 1)
    with pytest.raises(TypeError, match=expected):
        landlab.plot.imshow_grid(grid, values)
    with pytest.raises(TypeError, match=expected):
        landlab.plot.imshow_grid(grid, "foo")


@pytest.mark.parametrize("at", ["link", "patch", "corner", "face"])
def test_imshow_grid_unsupported_location(at):
    expected = (
        "value location, \\'[a-z]+\\', is not supported \\(must be "
        "one of 'node', 'cell'\\)"
    )
    grid = landlab.RasterModelGrid((5, 3))
    grid.add_zeros("z", at=at)
    with pytest.raises(TypeError, match=expected):
        landlab.plot.imshow_grid(grid, "z")
    with pytest.raises(TypeError, match=expected):
        landlab.plot.imshow_grid(grid, grid[at]["z"])


def test_values_at_is_deprecated():
    grid = landlab.RasterModelGrid((5, 3))
    grid.add_zeros("topographic__elevation", at="node")
    with pytest.deprecated_call(
        match="the 'values_at' keyword is deprecated, use the 'at' keyword instead"
    ):
        landlab.plot.imshow_grid(grid, "topographic__elevation", values_at="node")


@pytest.mark.parametrize("at", ["node", "link", "patch", "corner", "face", "cell"])
def test_guess_location(at):
    grid = landlab.RasterModelGrid((3, 4))
    values = grid.add_zeros("z", at=at)

    assert _guess_location_from_name(grid, "z") == at
    assert _guess_location_from_name(grid, "foo") is None

    guess = _guess_location_from_size(grid, np.empty_like(values))
    assert grid.number_of_elements(guess) == values.size
    assert _guess_location_from_size(grid, np.empty(grid.number_of_links + 1)) is None


@pytest.mark.parametrize("actual", ["node", "cell"])
@pytest.mark.parametrize("at", ["link", "patch", "corner", "face"])
def test_location_node_cell_first(actual, at):
    grid = landlab.RasterModelGrid((3, 4))
    values = grid.add_zeros("z", at=actual)
    grid.add_zeros("z", at=at)

    assert _guess_location_from_name(grid, "z") == actual
    assert _guess_location_from_size(grid, np.empty_like(values)) == actual



""" Implement the DualFramedVoronoiGraph

@author sebastien lenard
@date 2022, Aug
"""

from ..dual import DualGraph
from ..voronoi.dual_voronoi import DualVoronoiGraph
from .framed_voronoi import FramedVoronoiGraph


class DualFramedVoronoiGraph(DualGraph, FramedVoronoiGraph):

    """Graph of a unstructured grid of Voronoi Delaunay cells and
    irregular patches. It is a special type of VoronoiDelaunay graph in which
    the initial set of points is arranged in a fixed lattice (e.g. like a rectangular
    raster grid) named here "layout" and the core points are then moved aroung their
    initial position by a random distance, lower than a certain threshold.

    Examples
    --------
    >>> from landlab.graph import DualFramedVoronoiGraph

    >>> graph = DualFramedVoronoiGraph((3, 3), seed=200)
    >>> graph.number_of_nodes
    9

    >>> graph.x_of_node[2:4]    # doctest: +NORMALIZE_WHITESPACE
    array([ 2.,  0.])
    >>> graph.y_of_node[2:4]    # doctest: +NORMALIZE_WHITESPACE
    array([ 0.   ,  0.749])
    >>> graph.y_of_node[5]    # doctest: +NORMALIZE_WHITESPACE
    1.2509999999999999
    """

    def __init__(
        self,
        shape,
        xy_spacing=(1.0, 1.0),
        xy_of_lower_left=(0.0, 0.0),
        sort=False,
        xy_min_spacing=(0.5, 0.5),
        seed=200,
    ):
        """Create the graph.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of nodes.
        xy_spacing : float or tuple of float, optional
            Node spacing along x and y coordinates. If float, same spacing at x and y.
        xy_of_lower_left : tuple, optional
            Minimum *x*-of-node and *y*-of-node values. Depending on the grid,
            there may not be a node present at this location.
        sort: bool
            If ``True``, nodes, links and patches are re-numbered according to
            their position.
        xy_min_spacing: float or tuple of float, optional
            Final minimal spacing between nodes. Random moves of the core nodes
            around their position cannot be above this threshold:
            ``(xy_spacing - xy_min_spacing) / 2``
            If ``float``, same minimal spacing for *x* and *y*.
        seed: int, optional
            Seed used to generate the random *x* and *y* moves. When set,
            controls a pseudo-randomness of moves to ensure reproducibility.
            When ``None``, seed is random and the moves of coordinates are
            completely random.

        Returns
        -------
        DualFramedVoronoiGraph
            A newly-created graph.

        Examples
        --------
        Create a grid with 3 rows and 2 columns of nodes.

        >>> from landlab.graph import DualFramedVoronoiGraph
        >>> graph = DualFramedVoronoiGraph((3, 2), xy_spacing=1.0)
        >>> graph.number_of_nodes
        6
        """
        FramedVoronoiGraph.__init__(
            self,
            shape,
            xy_spacing=xy_spacing,
            xy_of_lower_left=xy_of_lower_left,
            sort=True,
            xy_min_spacing=xy_min_spacing,
            seed=seed,
        )

        DualVoronoiGraph.__init__(
            self,
            (self._y_of_node, self._x_of_node),
            perimeter_links=self._perimeter_links,
            sort=False,
        )

        if sort:
            self.sort()



#!/usr/bin/env python3
"""
Created on Mon Apr 30 09:17:36 2018

@author: barnhark
"""

import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.bmi import wrap_as_bmi
from landlab.components import LithoLayers


def test_litholayers_as_bmi():
    """Test Litholayers can be wrapped with a BMI."""
    wrap_as_bmi(LithoLayers)


def test_z0s_ids_different_shape():
    """Test that providing z0s and ids of different shapes raises an error."""
    mg = RasterModelGrid((3, 3))
    z0s = [-4, -3, -2, -1, 0, 1, 2, 3, 4]
    ids = [1, 2, 1, 2, 1, 2, 1, 2, 1, 2]
    attrs = {"K_sp": {1: 0.001, 2: 0.0001}}
    with pytest.raises(ValueError):
        LithoLayers(mg, z0s, ids, attrs)


def test_z0s_bad_order():
    """Test that providing z0s in a bad order raises an error."""
    mg = RasterModelGrid((3, 3))
    z0s = [-4, -3, -2, -1, 0, 1, 2, 6, 4]
    ids = [1, 2, 1, 2, 1, 2, 1, 2, 1]
    attrs = {"K_sp": {1: 0.001, 2: 0.0001}}
    with pytest.raises(ValueError):
        LithoLayers(mg, z0s, ids, attrs)


def test_bad_function():
    """Test that providing a function of three variables."""
    mg = RasterModelGrid((3, 3))
    z0s = [-4, -3, -2, -1, 0, 1, 2, 4, 6]
    ids = [1, 2, 1, 2, 1, 2, 1, 2, 1]
    attrs = {"K_sp": {1: 0.001, 2: 0.0001}}
    with pytest.raises(ValueError):
        LithoLayers(mg, z0s, ids, attrs, function=lambda x, y, z: 0 * x + 0 * y + z)


def test_function_returns_scalar():
    mg = RasterModelGrid((3, 3))
    z0s = [-4, -3, -2, -1, 0, 1, 2, 4, 6]
    ids = [1, 2, 1, 2, 1, 2, 1, 2, 1]
    attrs = {"K_sp": {1: 0.001, 2: 0.0001}}
    with pytest.raises(ValueError):
        LithoLayers(mg, z0s, ids, attrs, function=lambda x, y: 1.0)


def test_function_returns_wrong_number_of_values():
    mg = RasterModelGrid((3, 3))
    z0s = [-4, -3, -2, -1, 0, 1, 2, 4, 6]
    ids = [1, 2, 1, 2, 1, 2, 1, 2, 1]
    attrs = {"K_sp": {1: 0.001, 2: 0.0001}}
    with pytest.raises(ValueError):
        LithoLayers(mg, z0s, ids, attrs, function=lambda x, y: np.array([1.0]))



import numpy as np


def angle_finder(grid, dn, cn, rn):
    """Find the interior angle between two vectors on a grid.

    Parameters
    ----------
    grid : ModelGrid
        A landlab grid.
    dn : int or array of int
        Node or nodes at the end of the first vector.
    cn : int or array of int
        Node or nodes at the vertex between vectors.
    rn : int or array of int
        Node or nodes at the end of the second vector.

    Returns
    -------
    float or array of float
        Angle between vectors (in radians).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.lateral_erosion.node_finder import angle_finder

    >>> grid = RasterModelGrid((3, 4))
    >>> np.rad2deg(angle_finder(grid, 8, 5, 0))
    90.0
    >>> np.rad2deg(angle_finder(grid, (8, 9, 10, 6), 5, 6))
    array([ 135.,   90.,   45.,    0.])
    """
    vertex = np.take(grid.x_of_node, cn), np.take(grid.y_of_node, cn)
    vec_1 = [
        np.take(grid.x_of_node, dn) - vertex[0],
        np.take(grid.y_of_node, dn) - vertex[1],
    ]
    vec_2 = [
        np.take(grid.x_of_node, rn) - vertex[0],
        np.take(grid.y_of_node, rn) - vertex[1],
    ]

    return np.arccos(
        (vec_1[0] * vec_2[0] + vec_1[1] * vec_2[1])
        / np.sqrt((vec_1[0] ** 2 + vec_1[1] ** 2) * (vec_2[0] ** 2 + vec_2[1] ** 2))
    )


def forty_five_node(donor, i, receiver, neighbors, diag_neigh):
    radcurv_angle = 0.67

    lat_node = 0
    # In Landlab 2019: diagonal list goes [NE, NW, SW, SE]. Node list are ordered as [E,N,W,S]
    # if water flows SE-N OR if flow NE-S or E-NW or E-SW, erode west node
    if (
        donor == diag_neigh[0]
        and receiver == neighbors[3]
        or donor == diag_neigh[3]
        and receiver == neighbors[1]
        or donor == neighbors[0]
        and receiver == diag_neigh[2]
        or donor == neighbors[0]
        and receiver == diag_neigh[1]
    ):
        lat_node = neighbors[2]
    # if flow is from SW-N or NW-S or W-NE or W-SE, erode east node
    elif (
        donor == diag_neigh[1]
        and receiver == neighbors[3]
        or donor == diag_neigh[2]
        and receiver == neighbors[1]
        or donor == neighbors[2]
        and receiver == diag_neigh[3]
        or donor == neighbors[2]
        and receiver == diag_neigh[0]
    ):
        lat_node = neighbors[0]
    # if flow is from SE-W or SW-E or S-NE or S-NW, erode north node
    elif (
        donor == diag_neigh[3]
        and receiver == neighbors[2]
        or donor == diag_neigh[2]
        and receiver == neighbors[0]
        or donor == neighbors[3]
        and receiver == diag_neigh[0]
        or donor == neighbors[3]
        and receiver == diag_neigh[1]
    ):
        lat_node = neighbors[1]
    # if flow is from NE-W OR NW-E or N-SE or N-SW, erode south node
    elif (
        donor == diag_neigh[0]
        and receiver == neighbors[2]
        or donor == diag_neigh[1]
        and receiver == neighbors[0]
        or donor == neighbors[1]
        and receiver == diag_neigh[3]
        or donor == neighbors[1]
        and receiver == diag_neigh[2]
    ):
        lat_node = neighbors[3]
    return lat_node, radcurv_angle


def ninety_node(donor, i, receiver, link_list, neighbors, diag_neigh):
    # if flow is 90 degrees
    if donor in diag_neigh and receiver in diag_neigh:
        radcurv_angle = 1.37
        # if flow is NE-SE or NW-SW, erode south node
        if (
            donor == diag_neigh[0]
            and receiver == diag_neigh[3]
            or donor == diag_neigh[1]
            and receiver == diag_neigh[2]
        ):
            lat_node = neighbors[3]
        # if flow is SW-NW or SE-NE, erode north node
        elif (
            donor == diag_neigh[2]
            and receiver == diag_neigh[1]
            or donor == diag_neigh[3]
            and receiver == diag_neigh[0]
        ):
            lat_node = neighbors[1]
        # if flow is SW-SE or NW-NE, erode east node
        elif (
            donor == diag_neigh[2]
            and receiver == diag_neigh[3]
            or donor == diag_neigh[1]
            and receiver == diag_neigh[0]
        ):
            lat_node = neighbors[0]
        # if flow is SE-SW or NE-NW, erode west node
        elif (
            donor == diag_neigh[3]
            and receiver == diag_neigh[2]
            or donor == diag_neigh[0]
            and receiver == diag_neigh[1]
        ):
            lat_node = neighbors[2]
    elif donor not in diag_neigh and receiver not in diag_neigh:
        radcurv_angle = 1.37
        # if flow is from east, erode west node
        if donor == neighbors[0]:
            lat_node = neighbors[2]
        # if flow is from north, erode south node
        elif donor == neighbors[1]:
            lat_node = neighbors[3]
        # if flow is from west, erode east node
        elif donor == neighbors[2]:
            lat_node = neighbors[0]
        # if flow is from south, erode north node
        elif donor == neighbors[3]:
            lat_node = neighbors[1]
    return lat_node, radcurv_angle


def straight_node(donor, i, receiver, neighbors, diag_neigh):
    # ***FLOW LINK IS STRAIGHT, NORTH TO SOUTH***#
    if donor == neighbors[1] or donor == neighbors[3]:
        # print "flow is stright, N-S from ", donor, " to ", flowdirs[i]
        radcurv_angle = 0.23
        # neighbors are ordered E,N,W, S
        # if the west cell is boundary (neighbors=-1), erode from east node
        if neighbors[2] == -1:
            lat_node = neighbors[0]
        elif neighbors[0] == -1:
            lat_node = neighbors[2]
        else:
            # if could go either way, choose randomly. 0 goes East, 1 goes west
            ran_num = np.random.randint(0, 2)
            if ran_num == 0:
                lat_node = neighbors[0]
            if ran_num == 1:
                lat_node = neighbors[2]
    # ***FLOW LINK IS STRAIGHT, EAST-WEST**#
    elif donor == neighbors[0] or donor == neighbors[2]:
        radcurv_angle = 0.23
        #  Node list are ordered as [E,N,W,S]
        # if the north cell is boundary (neighbors=-1), erode from south node
        if neighbors[1] == -1:
            lat_node = neighbors[3]
        elif neighbors[3] == -1:
            lat_node = neighbors[1]
        else:
            # if could go either way, choose randomly. 0 goes south, 1 goes north
            ran_num = np.random.randint(0, 2)
            if ran_num == 0:
                lat_node = neighbors[1]
            if ran_num == 1:
                lat_node = neighbors[3]
    # if flow is straight across diagonal, choose node to erode at random
    elif donor in diag_neigh and receiver in diag_neigh:
        radcurv_angle = 0.23
        if receiver == diag_neigh[0]:
            poss_diag_nodes = neighbors[0 : 1 + 1]
        elif receiver == diag_neigh[1]:
            poss_diag_nodes = neighbors[1 : 2 + 1]
        elif receiver == diag_neigh[2]:
            poss_diag_nodes = neighbors[2 : 3 + 1]
        elif receiver == diag_neigh[3]:
            poss_diag_nodes = [neighbors[3], neighbors[0]]
        ran_num = np.random.randint(0, 2)
        if ran_num == 0:
            lat_node = poss_diag_nodes[0]
        if ran_num == 1:
            lat_node = poss_diag_nodes[1]
    return lat_node, radcurv_angle


def node_finder(grid, i, flowdirs, drain_area):
    """Find lateral neighbor node of the primary node for straight, 45 degree,
    and 90 degree channel segments.

    Parameters
    ----------
    grid : ModelGrid
        A Landlab grid object
    i : int
        node ID of primary node
    flowdirs : array
        Flow direction array
    drain_area : array
        drainage area array

    Returns
    -------
    lat_node : int
        node ID of lateral node
    radcurv_angle : float
        inverse radius of curvature of channel at lateral node
    """
    # receiver node of flow is flowdirs[i]
    receiver = flowdirs[i]

    # find indicies of where flowdirs=i to find donor nodes.
    # will donor nodes always equal the index of flowdir list?
    inflow = np.where(flowdirs == i)

    # if there are more than 1 donors, find the one with largest drainage area

    if len(inflow[0]) > 1:
        drin = drain_area[inflow]
        drmax = max(drin)
        maxinfl = inflow[0][np.where(drin == drmax)]
        # if donor nodes have same drainage area, choose one randomly
        if len(maxinfl) > 1:
            ran_num = np.random.randint(0, len(maxinfl))
            maxinfln = maxinfl[ran_num]
            donor = [maxinfln]
        else:
            donor = maxinfl
        # if inflow is empty, no donor
    elif len(inflow[0]) == 0:
        donor = i
    # else donor is the only inflow
    else:
        donor = inflow[0]
    # now we have chosen donor cell, next figure out if inflow/outflow lines are
    # straight, 45, or 90 degree angle. and figure out which node to erode
    link_list = grid.links_at_node[i]
    # this gives list of active neighbors for specified node
    # the order of this list is: [E,N,W,S]
    neighbors = grid.active_adjacent_nodes_at_node[i]
    # this gives list of all diagonal neighbors for specified node
    # the order of this list is: [NE,NW,SW,SE]
    diag_neigh = grid.diagonal_adjacent_nodes_at_node[i]
    angle_diff = np.rad2deg(angle_finder(grid, donor, i, receiver))

    if (donor == flowdirs[i]) or (donor == i):
        # this is a sink. no lateral ero
        radcurv_angle = 0.0
        lat_node = 0
    elif np.isclose(angle_diff, 0.0) or np.isclose(angle_diff, 180.0):
        [lat_node, radcurv_angle] = straight_node(
            donor, i, receiver, neighbors, diag_neigh
        )
    elif np.isclose(angle_diff, 45.0) or np.isclose(angle_diff, 135.0):
        [lat_node, radcurv_angle] = forty_five_node(
            donor, i, receiver, neighbors, diag_neigh
        )
    elif np.isclose(angle_diff, 90.0):
        [lat_node, radcurv_angle] = ninety_node(
            donor, i, receiver, link_list, neighbors, diag_neigh
        )
    else:
        lat_node = 0
        radcurv_angle = 0.0

    dx = grid.dx
    # INVERSE radius of curvature.
    radcurv_angle = radcurv_angle / dx
    return int(lat_node), radcurv_angle



#!/usr/bin/env python3
"""
Calculate dimensionless discharge of stream sections based on Tang et
al. (2019)
"""

import numpy as np
from scipy import constants

from landlab import Component
from landlab.utils.return_array import return_array_at_node


class DimensionlessDischarge(Component):
    r"""Component that calculates dimensionless discharge of stream
     segments.

    The dimensionless discharge model calculates the unitless  discharge
    value of streams and can be used to help determine locations of
    debris flows. It uses an equation from Tang et al. (2019) to
    calculate the dimensionless discharge as well as the threshold for
    whether a debris flow will occur for a specified location.

    .. math::

        q* = \frac{q}{\sqrt{\frac{\rho_s-\rho}{\rho}*g*D_{50}^3}}

     where :math:`q*` is the dimensionless discharge value for a stream
     segment, :math:`q` is flux in a stream segment,  :math:`\rho_s`
     is soil density, :math:`\rho` is water density, :math:`g` is the
     gravitational constant, and :math:`D_50` is the average sediment
     partical size in the stream segment area.

     Constants C and N are coefficients used in the slope-dependent
     equation

     .. math::

        q*_{thresold} = \frac{C}{(tan(\theta))^N}

     to determine whether the dimensionless discharge calculated
     exceeds thresholds for a sediment-laden (Upper Limit) or
     water-producing only (Lower Limit) debris flow.  C and N are
     empirically-derived by comparing dimensionless discharge estimates
     against previous debris flow events.  The values will vary based on
     local geology and soil type. In southern California values of C=12
     and N=0.85 (upper limits, Tang et al, 2019) and C=4.29, N=0.78
     (lower limits, Tang et al, 2019) have been used, while values of
     C=0.195 and N=1.27 have been in used in the Italian Dolomites
     (Gregoretti and Fontana, 2008). Default values are C=12 and N=0.85.

     Examples
     --------
     >>> from landlab.components import DimensionlessDischarge
     >>> from landlab import RasterModelGrid
     >>> import random
     >>> watershed_grid = RasterModelGrid((3, 3))
     >>> surface_water__unit_discharge = watershed_grid.add_ones(
     ...    "surface_water__unit_discharge", at="node"
     ... )
     >>> d50 = watershed_grid.add_ones(
     ...    "channel_bottom_sediment_grain__d50_diameter", at="node"
     ... )
     >>> watershed_grid.at_node ['topographic__elevation'] = \
     ... np.array([[1.1, 2, 3, 4, 2, 3, 4, 5, 3]])
     >>> dd = DimensionlessDischarge(watershed_grid, gravity = 9.8)
     >>> dd.run_one_step()
     >>> print(watershed_grid.at_node['dimensionless_discharge'])
     [ 0.55372743  0.55372743  0.55372743  0.55372743  0.55372743
             0.55372743  0.55372743  0.55372743  0.55372743]



     References
     ----------
     Tang, H., McGuire, L. A., Rengers, F. K., Kean, J. W., Staley,
     D. M., & Smith, J. B. (2019). Developing and Testing Physically
     Based Triggering Thresholds for Runoff-Generated Debris Flows.
     Geophysical Research Letters, 46(15), 8830–8839.
     https://doi.org/10.1029/2019GL083623

    """

    _name = "DimensionlessDischargeModel"

    _unit_agnostic = False

    _info = {
        "dimensionless_discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "none",
            "mapping": "node",
            "doc": "Dimensionless discharge value for a stream segment.",
        },
        "dimensionless_discharge_above_threshold": {
            "dtype": bool,
            "intent": "out",
            "optional": False,
            "units": "none",
            "mapping": "node",
            "doc": (
                "True if dimensionless discharge value is above threshold value, "
                "false otherwise."
            ),
        },
        "dimensionless_discharge_threshold": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "none",
            "mapping": "node",
            "doc": "Dimensionless discharge threshold for each stream segment.",
        },
        "surface_water__unit_discharge": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**2/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water per unit width",
        },
        "channel_bottom_sediment_grain__d50_diameter": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m",
            "mapping": "node",
            "doc": "soil grain size average in stream segment",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(
        self,
        grid,
        soil_density=1330,
        water_density=997.9,
        C=12.0,
        N=0.85,
        gravity=constants.g,
        channel_bottom_sediment_grain__d50_diameter="channel_bottom_sediment_grain__d50_diameter",
    ):
        """Initialize the DimensionlessDischarge.

        Parameters
        ----------
        soil_density : float, field name, or array, optional
            density of soil (kg/m^3) (defaults to 1330)
        water_density : float, optional
            density of water (kg/m^3) (defaults to 997.9)
        C : float, optional
            Numerator of the debris flow threshold equation; Empirically
            derived constant (See Tang et al. 2019). (defaults to 12.0)
        N : float, optional
            Exponent for slope in the denominator of the debris flow
            threshold equation; Empirically derived constant (See Tang
            et al. 2019). (defaults to 0.85)
        gravity : float, optional
            (defaults to scipy's standard acceleration of gravity)
        channel_bottom_sediment_grain__d50_diameter : float, field_name, or array
            defaults to the field name
            "channel_bottom_sediment_grain__d50_diameter"
        """

        super().__init__(grid)

        # Store parameters

        # scalar parameters.
        self._c = C
        self._n = N
        self._water_density = water_density
        self._gravity = gravity

        # get topographic__elevation values and change into slope of a
        # stream segment
        self._stream_slopes = self._elevationToSlope()

        # both soil density and d50 can be float, array, or field name.
        self._soil_density = return_array_at_node(self.grid, soil_density)
        self._channel_bottom_sediment_grain__d50_diameter = return_array_at_node(
            self.grid, channel_bottom_sediment_grain__d50_diameter
        )

        # create output fields
        self.grid.add_zeros("dimensionless_discharge", at="node")
        self.grid.add_full(
            "dimensionless_discharge_above_threshold", False, at="node", dtype=bool
        )
        self.grid.add_zeros("dimensionless_discharge_threshold", at="node", dtype=float)

        # calculate threshold values for each segment
        self._calc_threshold()

    def _calc_threshold(self):
        self.grid.at_node["dimensionless_discharge_threshold"] = self._c / (
            self._stream_slopes**self._n
        )

    def _elevationToSlope(self):
        return self.grid.calc_slope_at_node(elevs="topographic__elevation")

    def run_one_step(self):
        # update slopes
        self._stream_slopes = self._elevationToSlope()

        # recalculate threshold with new slopes.
        self._calc_threshold()

        self.grid.at_node["dimensionless_discharge"] = self.grid.at_node[
            "surface_water__unit_discharge"
        ] / np.sqrt(
            ((self._soil_density - self._water_density) / self._water_density)
            * self._gravity
            * (self._channel_bottom_sediment_grain__d50_diameter**3)
        )

        self.grid.at_node["dimensionless_discharge_above_threshold"] = (
            self.grid.at_node["dimensionless_discharge"]
            > self.grid.at_node["dimensionless_discharge_threshold"]
        )



#! /usr/bin/env python
"""
Unit tests for landlab.components.carbonate.carbonate_producer
"""
from numpy.testing import assert_allclose, assert_raises

from landlab import RasterModelGrid
from landlab.components import CarbonateProducer


def test_properties():
    grid = RasterModelGrid((3, 3))
    elev = grid.add_zeros("topographic__elevation", at="node")
    grid.add_field("sea_level__elevation", 0.0, at="grid")
    elev[:] = -1.0
    cp = CarbonateProducer(grid)
    assert cp.extinction_coefficient == 0.1
    cp.extinction_coefficient = 0.04
    assert_allclose(cp.calc_carbonate_production_rate()[4], 0.009998656)
    assert cp.max_carbonate_production_rate == 0.01
    cp.max_carbonate_production_rate = 0.015
    assert_allclose(cp.calc_carbonate_production_rate()[4], 0.014997984)
    assert cp.surface_light == 2000.0
    cp.surface_light = 2250.0
    assert_allclose(cp.calc_carbonate_production_rate()[4], 0.014999393)
    assert cp.saturating_light == 400.0
    cp.saturating_light = 50.0
    assert_allclose(cp.calc_carbonate_production_rate()[4], 0.015)
    assert cp.tidal_range == 0.0
    cp.tidal_range = 10.0
    assert_allclose(cp.calc_carbonate_production_rate()[4], 0.00824751)


def test_preexisting_field():
    grid = RasterModelGrid((3, 3))
    grid.add_zeros("topographic__elevation", at="node")
    grid.add_field("sea_level__elevation", 0.0, at="grid")
    grid.add_zeros("carbonate_thickness", at="node")
    cp = CarbonateProducer(grid)
    assert cp._carbonate_thickness is grid.at_node["carbonate_thickness"]


def test_exception_handling():
    grid = RasterModelGrid((3, 3))
    grid.add_zeros("topographic__elevation", at="node")
    grid.add_field("sea_level__elevation", 0.0, at="grid")
    assert_raises(ValueError, CarbonateProducer, grid, extinction_coefficient=0.0)
    assert_raises(ValueError, CarbonateProducer, grid, max_carbonate_production_rate=-1)
    assert_raises(ValueError, CarbonateProducer, grid, surface_light=-1)
    assert_raises(ValueError, CarbonateProducer, grid, saturating_light=-1)
    assert_raises(ValueError, CarbonateProducer, grid, tidal_range=-1)



from functools import cached_property

import numpy as np

from ...core.utils import as_id_array
from ...utils.decorators import read_only_array
from ..voronoi.voronoi import DelaunayGraph


class RadialGraphLayout:
    @staticmethod
    def number_of_nodes(shape):
        return np.sum(np.arange(1, shape[0] + 1)) * shape[1] + 1

    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_center=(0.0, 0.0)):
        """Create the node layout for a radial grid.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph.radial.radial import RadialGraphLayout
        >>> x, y = RadialGraphLayout.xy_of_node((1, 6))
        >>> x
        array([ 0. ,  1. ,  0.5, -0.5, -1. , -0.5,  0.5])
        >>> np.round(y / np.sin(np.pi / 3.0))
        array([ 0.,  0.,  1.,  1.,  0., -1., -1.])
        """
        n_rings, n_points = shape
        n_nodes = RadialGraphLayout.number_of_nodes(shape)

        x = np.empty((n_nodes,), dtype=float)
        y = np.empty((n_nodes,), dtype=float)

        x[0] = y[0] = 0.0
        offset = 1
        for ring in range(1, n_rings + 1):
            rho = spacing * ring
            d_theta = np.pi * 2 / (ring * shape[1])
            theta = np.arange(ring * shape[1]) * d_theta

            y[offset : offset + len(theta)] = rho * np.sin(theta)
            x[offset : offset + len(theta)] = rho * np.cos(theta)

            offset += len(theta)

        x = np.round(x, decimals=6)
        y = np.round(y, decimals=6)

        x += xy_of_center[0]
        y += xy_of_center[1]

        return (x, y)


class RadialGraphExtras:
    @property
    def shape(self):
        return self._shape

    @property
    def spacing(self):
        return self._spacing

    @property
    def origin(self):
        return self._origin

    @property
    def number_of_rings(self):
        return self.shape[0]

    @property
    def spacing_of_rings(self):
        return self.spacing

    @cached_property
    @read_only_array
    def radius_of_ring(self):
        return np.arange(0, self.number_of_rings, dtype=float) * self.spacing_of_rings

    @cached_property
    @read_only_array
    def angle_spacing_of_ring(self):
        return 2.0 * np.pi / self.nodes_per_ring

    @cached_property
    @read_only_array
    def nodes_per_ring(self):
        nodes_per_ring = np.empty(self.number_of_rings, dtype=int)
        nodes_per_ring[0] = 1
        nodes_per_ring[1:] = np.round(2.0 * np.pi * np.arange(1, self.number_of_rings))
        return nodes_per_ring

    @cached_property
    @read_only_array
    def ring_at_node(self):
        return np.repeat(np.arange(self.number_of_rings), self.nodes_per_ring)

    @cached_property
    @read_only_array
    def radius_at_node(self):
        return self.radius_of_ring[self.ring_at_node]

    @cached_property
    @read_only_array
    def angle_at_node(self):
        angle_at_node = np.empty(self.nodes_per_ring.sum(), dtype=float)
        angle_at_node[0] = 0.0
        offset = 1
        for n_nodes in self.nodes_per_ring[1:]:
            angles, step = np.linspace(
                0.0, 2 * np.pi, n_nodes, endpoint=False, retstep=True, dtype=float
            )
            angle_at_node[offset : offset + n_nodes] = np.add(
                angles, 0.5 * step, out=angles
            )
            offset += n_nodes
        return angle_at_node


class RadialGraph(RadialGraphExtras, DelaunayGraph):

    """Graph of a series of points on concentric circles.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import RadialGraph
    >>> graph = RadialGraph((1, 4), sort=True)
    >>> graph.number_of_nodes
    5
    >>> graph.y_of_node
    array([-1.,  0.,  0.,  0.,  1.])
    >>> graph.x_of_node
    array([ 0., -1.,  0.,  1.,  0.])
    """

    def __init__(self, shape, spacing=1.0, xy_of_center=(0.0, 0.0), sort=False):
        """Create a structured grid of triangles arranged radially.

        Parameters
        ----------
        shape : tuple of int
            Shape of the graph as number of rings and number of points
            in the first ring.
        spacing : float, optional
            Spacing between rings.
        xy_of_center : tuple of float, optional
            Coordinates of the node at the center of the grid.
        """
        try:
            spacing = float(spacing)
        except TypeError as exc:
            raise TypeError("spacing must be a float") from exc

        xy_of_center = tuple(np.broadcast_to(xy_of_center, 2))

        x_of_node, y_of_node = RadialGraphLayout.xy_of_node(
            shape, spacing=spacing, xy_of_center=xy_of_center
        )

        self._ring_spacing = spacing
        self._shape = tuple(shape)
        self._xy_of_center = xy_of_center

        DelaunayGraph.__init__(self, (y_of_node, x_of_node))

        if sort:
            self.sort()

    @property
    def xy_of_center(self):
        return self._xy_of_center

    @property
    def number_of_rings(self):
        """Number of node rings in grid.

        Returns
        -------
        int
            The number of node rings in the radial grid (not counting the
            center node).

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import RadialGraph
        >>> graph = RadialGraph((1, 4))
        >>> graph.number_of_rings
        1

        :meta landlab: info-grid
        """
        return self._shape[0]

    @property
    def spacing_of_rings(self):
        """Fixed distance between rings.

        Returns
        -------
        ndarray of float
            The distance from the center node of each node.

        >>> from landlab.graph import RadialGraph
        >>> graph = RadialGraph((2, 6), spacing=2.)
        >>> graph.spacing_of_rings
        2.0

        :meta landlab: info-grid, quantity
        """
        return self._ring_spacing

    @cached_property
    def radius_at_node(self):
        """Distance for center node to each node.

        Returns
        -------
        ndarray of float
            The distance from the center node of each node.

        >>> from landlab.graph import RadialGraph
        >>> graph = RadialGraph((2, 6), sort=True)
        >>> np.round(graph.radius_at_node, 3)
        array([ 2.,  2.,  2.,  2.,  2.,  1.,  1.,  2.,  1.,  0.,  1.,  2.,  1.,
                1.,  2.,  2.,  2.,  2.,  2.])

        :meta landlab: info-node, quantity
        """
        return np.sqrt(
            np.square(self.x_of_node - self._xy_of_center[0])
            + np.square(self.y_of_node - self._xy_of_center[1])
        )

    @cached_property
    def number_of_nodes_in_ring(self):
        """Number of nodes in each ring.

        Returns
        -------
        ndarray of int
            Number of nodes in each ring, excluding the center node.

        >>> from landlab.graph import RadialGraph
        >>> graph = RadialGraph((4, 6))
        >>> graph.number_of_nodes_in_ring
        array([ 6, 12, 24, 48])

        :meta landlab: info-node, quantity
        """
        return as_id_array(self._shape[1] * 2 ** np.arange(self.number_of_rings))



import numpy as np

cimport cython
cimport numpy as np


cdef extern from "math.h":
    double exp(double x) nogil

from libc.math cimport exp, log

DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t

DTYPE_INT = int
ctypedef np.int_t DTYPE_INT_t

DTYPE_UINT8 = np.uint8
ctypedef np.uint8_t DTYPE_UINT8_t

from libc.stdio cimport printf


def _sequential_ero_depo(np.ndarray[DTYPE_INT_t, ndim=1] stack_flip_ud_sel,
                    np.ndarray[DTYPE_INT_t, ndim=1] flow_receivers,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] cell_area,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] q,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] qs,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] qs_in,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] Es,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] Er,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] Q_to_the_m,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] slope,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] H,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] br,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] sed_erosion_term,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] bed_erosion_term,
                    np.ndarray[DTYPE_FLOAT_t, ndim=1] K_sed,
                    DTYPE_FLOAT_t v,
                    DTYPE_FLOAT_t phi,
                    DTYPE_FLOAT_t F_f,
                    DTYPE_FLOAT_t H_star,
                    DTYPE_FLOAT_t dt,
                    DTYPE_FLOAT_t thickness_lim):

    """Calculate and qs and qs_in."""
    # define internal variables
    cdef unsigned int node_id
    cdef unsigned int i
    cdef double H_Before
    cdef double vol_SSY_riv
    vol_SSY_riv =0.0


    for node_id in stack_flip_ud_sel:
        qs_out = (qs_in[node_id] + Es[node_id]*cell_area[node_id] + (1.0-F_f)*Er[node_id]* cell_area[node_id]) / \
                        (1.0+(v*cell_area[node_id]/q[node_id]))
        depo_rate = v*qs_out/q[node_id]
        H_loc       =   H[node_id]
        H_Before    =   H[node_id]
        slope_loc   =   slope[node_id]
        sed_erosion_loc = sed_erosion_term[node_id]
        bed_erosion_loc = bed_erosion_term[node_id]

        # Correct for thick soils where soil thickness can grow to inf
        if (H_loc > thickness_lim or slope_loc <= 0 or   sed_erosion_loc==0):
            H_loc += (depo_rate / (1 - phi) - sed_erosion_loc/ (1 - phi)) * dt
        else:
            # Blowup
            if (depo_rate == (K_sed[node_id] * Q_to_the_m[node_id] * slope_loc)) :
                H_loc = H_loc * log(
                    ((sed_erosion_loc/ (1 - phi)) / H_star)
                    * dt
                    + exp(H_loc / H_star)
                )
            # No blowup
            else:
                H_loc = H_star* np.log(
                    (1 / ((depo_rate / (1 - phi)) / (sed_erosion_loc/ (1 - phi))- 1))
                    * (
                        exp((depo_rate / (1 - phi)- (sed_erosion_loc/ (1 - phi)))* (dt / H_star))
                        * (((depo_rate/ (1 - phi)/ (sed_erosion_loc/ (1 - phi))) - 1)* exp(H_loc/ H_star)+1)
                        - 1
                        )
                )
            # In case soil depth evolves to infinity, fall back to no entrainment
            if H_loc == np.inf:
                H_loc =H[node_id]+ (depo_rate / (1 - phi) - sed_erosion_loc/ (1 - phi)) * dt


        H_loc = max(0,H_loc)
        ero_bed = bed_erosion_loc* (exp(-H_loc / H_star))
        qs_out_adj =  qs_in[node_id] - ((H_loc - H_Before)*(1-phi)*cell_area[node_id]/dt) +(1.0-F_f)*ero_bed* cell_area[node_id]# should always be bigger than 0

        qs[node_id] = qs_out_adj
        qs_in[node_id] = 0
        qs_in[flow_receivers[node_id]] += qs[node_id]

        H[node_id] = H_loc
        br[node_id]  += -dt * ero_bed
        vol_SSY_riv += F_f*ero_bed* cell_area[node_id]
    return vol_SSY_riv



import test_breach_c as breach


def test_priority_queue():
    breach.test_priority_queue()


def test_init_flow_direction_queues():
    breach.test_init_flow_direction_queues()


def test_set_flooded_and_outlet():
    breach.test_set_flooded_and_outlet()


def test_set_receiver():
    breach.test_set_receiver()


def test_set_donor_properties():
    breach.test_set_donor_properties()


def test_direct_flow_c():
    breach.test_direct_flow_c()


def test_direct_flow():
    breach.test_direct_flow()



from landlab import Component
from landlab.grid.network import NetworkModelGrid


class SedimentPulserBase(Component):

    """Base class of :class:`~.SedimentPulserAtLinks` and :class:`~.SedimentPulserEachParcel`.

    :class:`~.SedimentPulserAtLinks` and :class:`~.SedimentPulserEachParcel` run the
    landlab :class:`~.DataRecord` :meth:`~.DataRecord.add_item` method on a
    :class:`~.DataRecord` configured for :class:`~.NetworkSedimentTransporter`.


    .. codeauthor: Jeff Keck, Allison Pfeiffer, Shelby Ahrendt
                   (with help from Eric Hutton and Katy Barnhart)

    Parameters
    ----------
    grid : ModelGrid
        landlab *ModelGrid* to place sediment parcels on.
    parcels: landlab DataRecord
        Tracks parcel location and variables
    D50: float, optional
        median grain size [m]
    D84_D50: float, optional
        ratio of 84th percentile grain size to the median grain size
    rho_sediment : float, optional
        Sediment grain density [kg / m^3].
    parcel_volume : float, optional
        parcel volume used for all parcels that do not have a specified volume
    abrasion_rate: float, optional
        volumetric abrasion exponent [1/m]


    Examples
    --------
    >>> import numpy as np
    >>> from landlab import NetworkModelGrid

    >>> y_of_node = (0, 100, 200, 200, 300, 400, 400, 125)
    >>> x_of_node = (0, 0, 100, -50, -100, 50, -150, -100)
    >>> nodes_at_link = ((1, 0), (2, 1), (1, 7), (3, 1), (3, 4), (4, 5), (4, 6))
    >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
    >>> grid.at_link["channel_width"] = np.full(grid.number_of_links, 1.0)  # m
    >>> grid.at_link["channel_slope"] = np.full(grid.number_of_links, .01)  # m / m
    >>> grid.at_link["reach_length"] = np.full(grid.number_of_links, 100.0)  # m
    >>> make_pulse_base = SedimentPulserBase(grid)
    >>> make_pulse_base._parcels

    SedimentPulserBase does not have any methods for adding a pulse

    >>> a_pulse = make_pulse_base()
    Traceback (most recent call last):
    ...
    NotImplementedError: the base component has no call method
    """

    _name = "SedimentPulserBase"

    _unit_agnostic = False

    _info = {}  # works with the DataRecord

    def __init__(
        self,
        grid,
        parcels=None,
        D50=0.05,
        D84_D50=2.1,
        rho_sediment=2650.0,
        parcel_volume=0.5,
        abrasion_rate=0.0,
    ):
        self._grid = grid
        self._parcels = parcels
        self._D50 = D50
        self._D84_D50 = D84_D50
        self._rho_sediment = rho_sediment
        self._parcel_volume = parcel_volume
        self._abrasion_rate = abrasion_rate

        if not isinstance(grid, NetworkModelGrid):
            raise ValueError(
                "NetworkSedimentTransporter: grid must be NetworkModelGrid"
            )

    def __call__(self):
        """__call__ is not implemented for this component."""
        raise NotImplementedError("the base component has no call method")
